import sys
import tkinter as tk
import tkinter.ttk as ttk
from tkinter.scrolledtext import ScrolledText
from typing import Any

from .base import ViewPage
from .selection_page import SelectionPage


class RunPage(ViewPage):
    """Run Page of the PyCurator View component.

    Parameters
    ----------
    *args : tuple, optional
        Positional arguments for ViewPage constructor, which further
        passes to ttk.Frame.
    **kwargs : dict, optional
        Keyword arguments for ViewPage constructor, which further
        passes to ttk.Frame.

    See Also
    --------
    pycurator.gui.base.ViewPage
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.prev_msg = None

        # Create global runtime frames/widgets
        self.output_frame = ttk.Frame(self)
        self.runtime_output = tk.scrolledtext.ScrolledText(
            master=self.output_frame,
        )
        self.runtime_output.config(state='disabled')
        self.progress_bar = ttk.Progressbar(master=self.output_frame)

        self.progress_text_frame = ttk.Frame(master=self)
        self.progress_label_placeholder = ttk.Label(
            master=self.progress_text_frame,
            text='Querying:'
        )
        self.progress_determinate_num = ttk.Label(
            master=self.progress_text_frame
        )
        self.progress_label = ttk.Label(master=self.progress_text_frame)

        self.button_frame = ttk.Frame(master=self)
        self.back_button = ttk.Button(
            master=self.button_frame,
            text='Back',
            command=lambda: self.controller.show(SelectionPage),
            state='disabled'
        )
        self.stop_button = ttk.Button(
            master=self.button_frame,
            text='Stop',
            command=lambda: self.controller.request_execution()
        )
        self.exit_button = ttk.Button(
            master=self.button_frame,
            text='Exit',
            command=sys.exit
        )

    @ViewPage.no_overwrite
    def show(self) -> None:
        """Display runtime status updates and navigation buttons.

        Organizes the elements of the RunPage display. The output contains a
        ScrolledText widget for displaying ThreadedRun collector updates, a
        ProgressBar widget for updates on loop events, and Button widgets for
        paging back after completion, stopping the collector object during a
        run, and exiting the application upon collector termination/completion.

        Notes
        -----
        The ScrolledText widget updates variably using the self.process_queue
        function to retrieve runtime updates from the ThreadedRun collector
        object and self._update_output to push the updates to the display.

        The ProgressBar is used for loop queries in the collector object in
        order to provide continual updates to the user without pushing updates
        to the ScrolledText widget for every query made, as many of the APIs
        being querying may return hundreds to thousands of results to query.
        The indeterminate mode is used for repository queries performed in a
        while loop (such as Dryad and Zenodo), while determinate mode is used
        for repository queries over a fixed list of queries in a for loop
        (such as UCI searching over scraped pages).

        The navigation Buttons are set up in order to provide different
        functionality based on collector status. During runtime, the back
        button is disabled, and then reactivated upon collector completion. The
        end/exit button is set to stop the collector if pressed during runtime,
        and set to exit the program if pressed after collector completion.
        """

        # Arrange elements
        self.runtime_output.pack(side='top', expand=True, fill='both')

        self.progress_label_placeholder.pack(side='left', anchor='w')
        self.progress_determinate_num.pack(side='left', anchor='w')
        self.progress_label.pack(side='left', fill='x')
        self.progress_bar.pack(
            side='bottom',
            anchor='s',
            fill='x',
            expand=True
        )

        self.back_button.pack(side='left', fill='both')
        self.stop_button.pack(side='left', fill='both')

        self.button_frame.pack(side='bottom', expand=True)
        self.progress_text_frame.pack(side='bottom', fill='x')
        self.output_frame.pack(side='bottom')

        self.grid(row=0, column=0, sticky='nsew')
        self.tkraise()

        self.controller.run_collector()

    def reset_frame(self) -> None:
        """Modify frame widgets for reuse."""
        # Deactivate back button
        self.back_button.config(state='disabled')

        # Replace Exit button with Stop button
        self.exit_button.pack_forget()
        self.stop_button.pack(side='left')
        self.controller.run_collector()



from .Dataverse import DataverseCollector
from .Figshare import FigshareCollector
from .PapersWithCode import PapersWithCodeCollector

# Try to load Kaggle
try:
    from .Kaggle import KaggleCollector
except (OSError, ModuleNotFoundError):
    pass



from typing import Any, TypeVar, Union

from pandas import DataFrame

Page = TypeVar('Page', bound='gui.ViewPage')

SearchTerm = TypeVar('SearchTerm', bound=str)
SearchType = TypeVar('SearchType', bound=str)
SearchTuple = tuple[SearchTerm, SearchType]
SearchQuery = Union[SearchType, SearchTerm, SearchTuple]

SearchResult = Union[DataFrame, None]
TermResultDict = dict[SearchTerm, SearchResult]
TypeResultDict = dict[SearchType, SearchResult]
TermTypeResultDict = dict[SearchTuple, SearchResult]
QueryResultDict = dict[SearchQuery, SearchResult]

JSONDict = dict[str, Any]

AttributeKey = TypeVar('AttributeKey', bound=str)
AttributeValue = TypeVar('AttributeValue', bound=str)
AttributeDict = dict[AttributeKey, AttributeValue]

TKVarValue = TypeVar('TKVarValue', str, int, float, bool)



import warnings
from collections.abc import Collection
from typing import Any, Optional, Iterable, NoReturn

import pandas as pd

from pycurator._typing import SearchTerm
from pycurator.collectors import BaseCollector, BaseTermCollector


class ZenodoCollector(BaseTermCollector):
    """Zenodo collector for search term queries.

    Parameters
    ----------
    search_terms : list-like, optional
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        credentials: Optional[bool] = None
    ) -> None:
        super().__init__(
            repository_name='zenodo',
            search_terms=search_terms,
            credentials=credentials
        )
        self.base_url = 'https://zenodo.org/api/records'

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @BaseTermCollector.validate_search_term
    @BaseCollector._pb_indeterminate
    def get_individual_search_output(
            self,
            search_term: SearchTerm
    ) -> pd.DataFrame:
        """Returns information about all records from Zenodo.

        Parameters
        ----------
        search_term : str

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.

        Warns
        -----
        RuntimeWarning
            Unsuccessful query response from the API.
        """

        search_year = 2022
        search_df = pd.DataFrame()
        start_date = f'{search_year}-01-01'
        end_date = f'{search_year}-12-31'

        search_params = {
            'q': f'{search_term} AND created:[{start_date} TO {end_date}]',
            'page': 1,
            'size': 1000
        }

        response, output = self.get_request_output_and_update_query_ref(
            url=self.base_url,
            params=search_params,
            search_term=search_term,
            year=search_year,
            page=search_params['page']
        )

        # Handle any potential errors
        if response.status_code != 200:
            warnings.warn(
                f'{response.status_code}: Returning without results.',
                RuntimeWarning
            )
            self.continue_running = False
            self.terminate()

        while output.get('hits').get('total'):
            while (
                response.status_code == 200 and
                output.get('hits').get('hits')
            ):
                output = output['hits']['hits']

                output_df = pd.DataFrame(output)
                output_df['page'] = search_params['page']

                search_df = pd.concat(
                    [search_df, output_df]
                ).reset_index(drop=True)

                search_params['page'] += 1
                response, output = \
                    self.get_request_output_and_update_query_ref(
                        url=self.base_url,
                        params=search_params,
                        search_term=search_term,
                        year=search_year,
                        page=search_params['page']
                    )

            search_year -= 1
            start_date = f'{search_year}-01-01'
            end_date = f'{search_year}-12-31'

            search_params['q'] = \
                f'{search_term} AND created:[{start_date} TO {end_date}]'
            search_params['page'] = 1

            response, output = self.get_request_output_and_update_query_ref(
                url=self.base_url,
                params=search_params,
                search_term=search_term,
                year=search_year,
                page=search_params['page']
            )

        self.num_queries = False

        return search_df

    def get_query_metadata(self, object_paths: Iterable[Any]) -> NoReturn:
        raise NotImplementedError(
            'Zenodo does not provide object metadata.'
        )



import ast
from collections.abc import Collection
from typing import Optional, Union

import openml
import pandas as pd

from pycurator._typing import (
    SearchType,
    TypeResultDict
)
from pycurator.collectors import BaseTypeCollector
from pycurator.utils import validate_metadata_parameters


class OpenMLCollector(BaseTypeCollector):
    """OpenML collector for search type queries.

    Parameters
    ----------
    search_types : list-like, optional
        Types to search over. Can be (re)set via set_search_types() or
        passed in directly to search functions.
    credentials : str, optional (default=None)
        OpenML API key.
    """

    def __init__(
        self,
        search_types: Optional[Collection[SearchType]] = None,
        credentials: Optional[str] = None
    ) -> None:

        super().__init__(
            repository_name='openml',
            search_types=search_types
        )

        self.base_url = None

        if not openml.config.apikey:
            openml.config.apikey = credentials

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ('datasets', 'runs', 'tasks', 'evaluations')

    def _get_evaluations_search_output(self) -> pd.DataFrame:
        """Retrieve information on OpenML Evaluations."""
        evaluations_measures = openml.evaluations.list_evaluation_measures()
        evaluations_df = pd.DataFrame()

        # Get evaluation data for each available measure
        for measure in evaluations_measures:
            evaluations_dict = dict(
                openml.evaluations.list_evaluations(measure)
            )

            # Convert string list array_data to list
            if evaluations_dict.get('array_data'):
                try:
                    evaluations_dict['array_data'] = ast.literal_eval(
                        evaluations_dict['array_data']
                    )
                # Occurs if 'array_data' is already a list
                except ValueError:
                    pass

            measure_df = pd.DataFrame(
                map(
                    lambda evaluation: evaluation.__dict__,
                    evaluations_dict.values()
                )
            )

            evaluations_df = pd.concat(
                [evaluations_df, measure_df]
            ).reset_index(drop=True)

        return evaluations_df

    @BaseTypeCollector.validate_search_type
    def get_individual_search_output(
            self,
            search_type: SearchType
    ) -> pd.DataFrame:
        """Queries OpenML API for the specified search type.

        Parameters
        ----------
        search_type : {'datasets', 'runs', 'tasks', 'evaluations'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        ValueError
            Incorrect search_type provided.
        """

        self.status_queue.put(f'Querying {search_type}...')

        if search_type == 'evaluations':
            return self._get_evaluations_search_output()

        # Use query type to get necessary openml api functions
        base_command = getattr(openml, search_type)
        list_queries = getattr(base_command, f'list_{search_type}')

        index = 0
        size = 10000
        search_df = pd.DataFrame()

        self._update_query_ref(page=index)
        search_results = list_queries(offset=(index * size), size=size)

        while search_results:
            output_df = pd.DataFrame(search_results).transpose()
            output_df['page'] = index + 1
            search_df = pd.concat(
                [search_df, output_df]
            ).reset_index(drop=True)

            index += 1

            self._update_query_ref(page=index)
            search_results = list_queries(offset=(index * size), size=size)

        self.status_queue.put(f'{search_type} search complete.')

        return search_df

    @BaseTypeCollector.validate_search_type
    def get_query_metadata(
            self,
            object_paths: Union[str, Collection[str]],
            search_type: SearchType
    ) -> pd.DataFrame:
        """Retrieves the metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or list-like
        search_type : {'datasets', 'runs', 'tasks', 'evaluations'}

        Returns
        -------
        metadata_df : pandas.DataFrame

        Raises
        ------
        ValueError
            Invalid search_type provided.
        """

        object_paths = validate_metadata_parameters(object_paths)

        base_command = getattr(openml, search_type)
        get_query = getattr(base_command, f'get_{search_type[:-1:]}')

        queries = []
        error_queries = []
        for object_path in object_paths:
            self._update_query_ref(object_name=object_path)
            try:
                query = get_query(object_path).__dict__
                queries.append(query)
            except openml.exceptions.OpenMLServerException:
                error_queries.append(object_path)

        metadata_df = pd.DataFrame(queries)

        self.status_queue.put(f' {search_type} metadata query complete.')

        return metadata_df

    def get_all_metadata(
            self,
            search_dict: TypeResultDict
    ) -> TypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict
        """

        metadata_dict = dict()

        for query, df in search_dict.items():
            self.status_queue.put(f'Querying {query} metadata...')
            if query == 'datasets':
                id_name = 'did'
            elif query == 'runs':
                id_name = 'run_id'
            elif query == 'tasks':
                id_name = 'tid'
            else:
                raise ValueError(
                    f'Query \'{query}\' is not a valid search_type.'
                )

            object_paths = df[id_name].values

            metadata_dict[query] = self.get_query_metadata(
                object_paths=object_paths,
                search_type=query
            )

        self.status_queue.put('Metadata query complete.')

        return metadata_dict



from __future__ import annotations

import _tkinter
import os
import queue
import tkinter
import tkinter as tk
import tkinter.ttk as ttk
from typing import Optional, ParamSpec, Type, TypeVar, Union

from pycurator import collectors
from pycurator._typing import AttributeKey, AttributeValue, TKVarValue
from .base import ThreadedRun, ViewPage
from .landing_page import LandingPage
from .run_page import RunPage
from .selection_page import SelectionPage

Page = TypeVar('Page', bound=ViewPage)
P = ParamSpec('P')


class CuratorView(ttk.Frame):
    """View for the PyCurator UI.

    Parameters
    ----------
    parent : tk.Tk

    Attributes
    ----------
    pages : dict[str, ViewPage]
        Map from page name to page object.
    current_page : ViewPage

    See Also
    --------
    tk.Frame
    tk.Tk
    """

    def __init__(self, parent: tk.Tk) -> None:
        super().__init__(parent)
        self.pages = dict()
        self.current_page = None

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid(row=0, column=0, sticky='nsew')

        for F in (LandingPage, SelectionPage, RunPage):
            self.pages[F] = F(self)

    def set_controller(self, controller: CuratorController) -> None:
        """Setter for UI Controller of View object."""
        for page in self.pages.values():
            page.set_controller(controller)

    def show(self, page: Type[Page] = LandingPage) -> None:
        """Display the provided page on the UI."""
        self.current_page = self.pages[page]
        self.current_page.show()


class CollectorModel:
    """Model for the PyCurator UI. Extends the Collector class.

    Parameters
    ----------
    collector_class : Inherited BaseCollector class
    collector_name : str

    Attributes
    ----------
    collector_class : Inherited BaseCollector class
    collector_name : str
    collector : Inherited BaseCollector instance
    run_thread : threading.Thread
        Separate thread for data collection so that UI functionality
        is not interrupted.
    """

    def __init__(
            self,
            collector_class: Type[collectors.BaseCollector],
            collector_name: str
    ) -> None:
        self.collector_class = collector_class
        self.collector_name = collector_name
        self.collector = None
        self.run_thread = None

        self.requirements = {
            'search_terms': issubclass(
                self.collector_class,
                collectors.TermQueryMixin
            ),
            'search_types': issubclass(
                self.collector_class,
                collectors.TypeQueryMixin
            )
        }

    def initialize_collector(self, **param_val_kwargs: P.kwargs) -> None:
        """Instantiate Collector object from class and provided kwargs.

        Parameters
        ----------
        **param_val_kwargs : dict, optional
            Parameters for Collector initialization. See individual
            Collector classes for specifics.

        See Also
        --------
        pycurator.collectors : Classes for repository date collection.
        """

        self.collector = self.collector_class(**param_val_kwargs)

    def initialize_thread(self, thread_kwargs: P.kwargs) -> None:
        """Instantiate Collector Thread with runtime arguments.

        Parameters
        ----------
        thread_kwargs : dict, optional
            Parameters to provide to run function of collector.

        See Also
        -------
        pycurator.gui.base.ThreadedRun :
            Collector wrapper allowing for seamless UI performance by
            shifting collector actions to separate thread.
        """

        self.run_thread = ThreadedRun(
            collector=self.collector,
            kwargs=thread_kwargs
        )


class CuratorController:
    """Controller for the PyCurator UI.

    Parameters
    ----------
    model : CollectorModel, optional (default=None)
        Model for the PyCurator UI.
    view : CuratorView, optional (default=None)
        View for the PyCurator UI.

    Attributes
    ----------
    model : CollectorModel
    view : CuratorView
    runtime_param_vars : dict of parameter name to parameter value

    See Also
    --------
    CollectorModel
    CuratorView
    """

    def __init__(
            self,
            model: Optional[CollectorModel] = None,
            view: Optional[CuratorView] = None
    ) -> None:
        self.model = model
        self.view = view

        # Create placeholder for runtime_requirements property attribute
        self._runtime_requirements = None

        self.runtime_param_vars = dict()

    def request_next_page(self, *args: P.args) -> None:
        """Trigger the UI button corresponding to advancing pages."""
        try:
            self.view.current_page.next_page_button.invoke()
        except AttributeError:
            pass

    def request_execution(self) -> None:
        """Propagate Collector execution from UI to Collector."""
        self.model.collector.request_execution()

    def show(self, page: Type[Page]) -> None:
        """Show the provided page.

        Helper function for UI elements to request a page be shown
        without accessing the view component directly.

        Parameters
        ----------
        page : ViewPage
        """

        self.view.show(page)

    def set_model(
            self,
            collector: Type[collectors.BaseCollector],
            collector_name: str
    ) -> None:
        """Dynamically set the model component and helper variables.

        Parameters
        ----------
        collector : Inherited BaseCollector Class
        collector_name : str
        """

        self.model = CollectorModel(collector, collector_name)
        self.runtime_requirements = self.model.requirements
        self.runtime_param_vars[collector_name] = dict()

    def add_run_parameter(
            self,
            param: AttributeKey,
            value: AttributeValue
    ) -> None:
        """Setter for runtime_param_vars dict.

        Parameters provided are dynamically linked to the Collector
        currently loaded by the model. This allows values to be cached
        for later use, regardless of which Collectors are ran.

        Parameters
        ----------
        param : str
            Key for storing parameter value.
        value
        """

        self.runtime_param_vars[self.model.collector_name][param] = value

    def get_run_parameter(self, param: AttributeKey) -> AttributeValue:
        """Getter for runtime_param_vars dict of the current Collector."""
        return self.runtime_param_vars[self.model.collector_name][param]

    @property
    def runtime_requirements(self) -> dict[str, bool]:
        """Getter for requirements of the model's current Collector."""
        return self._runtime_requirements

    @runtime_requirements.setter
    def runtime_requirements(self, requirements: dict[str, bool]) -> None:
        """Setter for requirements of the model's current Collector."""
        self._runtime_requirements = requirements

    def evaluate_parameter(
            self,
            param_name: str
    ) -> Union[list[TKVarValue], TKVarValue, str]:
        """Retrieves values from the runtime_param_vars.

        Parameters
        ----------
        param_name : str

        Returns
        -------
        TKVarValue or Any
            For tkinter variables assigned to a parameter, returns the
            value from the tkinter.Variable.get() method.
            For non-tkinter variables, returns the variable itself.
        """

        param_var = self.get_run_parameter(param_name)

        if isinstance(param_var, dict):
            return [key for key, val in param_var.items() if val.get()]
        elif isinstance(param_var, tkinter.Variable):
            return param_var.get()
        else:
            return param_var

    def parse_run_parameters(self) -> None:
        """Ensure necessary requirements are present before running.

        See Also
        --------
        pycurator.collectors.base
        """

        param_val_kwargs = {
            param_name: self.evaluate_parameter(param_name)
            for param_name in self.runtime_param_vars[
                self.model.collector_name
            ]
        }

        # Add default save directory
        if not param_val_kwargs.get('save_dir'):
            param_val_kwargs['save_dir'] = os.path.join(
                'data',
                self.model.collector_name
            )
            os.makedirs(param_val_kwargs['save_dir'], exist_ok=True)

        missing_reqs = [
            param_name
            for param_name, is_req in self.runtime_requirements.items()
            if is_req and not param_val_kwargs.get(param_name)
        ]

        if missing_reqs:
            self.view.pages[SelectionPage].alert_missing_reqs(missing_reqs)
        else:
            self.initialize_run(**param_val_kwargs)

    def initialize_run(self, **param_val_kwargs: P.kwargs) -> None:
        """Initialize the model Collector and Thread for collection run.

        Parameters
        ----------
        **param_val_kwargs : dict, optional
            Initialization and runtime parameters for the collector
            model. save_dir and save_type are removed from the dict and
            passed to the thread initialization, while the rest are
            passed to the collector initialization.

        See Also
        -------
        CollectorModel
        pycurator.gui.base.ThreadedRun
        """

        save_dir = param_val_kwargs.pop('save_dir')
        save_type = param_val_kwargs.pop('save_type')

        # Set up the run thread
        self.model.initialize_collector(**param_val_kwargs)
        self.model.initialize_thread(
            {
                'save_dir': save_dir,
                'save_type': save_type,
            }
        )

        self.view.show(RunPage)

    def run_collector(self) -> None:
        """Start collection and propagate updates to UI."""
        self.model.run_thread.start()
        self.view.after(100, self.process_runtime_updates)

    def process_runtime_updates(self) -> None:
        """Push status update from collector queue to output."""
        # Check for updates to progress bar
        if self.model.run_thread.collector.num_queries is not None:
            self._update_progress_bar()

        # Get next object in queue and push to output widget.
        try:
            msg = self.model.run_thread.collector.status_queue.get_nowait()

            self._update_runtime_output(msg)

            # Check if process still running
            if self.model.collector.continue_running:
                self.view.after(100, self.process_runtime_updates)
            else:
                # Empty queue
                while not self.model.run_thread.collector.status_queue.empty():
                    self.process_runtime_updates()

                # Remove query message
                self.view.current_page.progress_label['text'] = 'Complete'

                # Stop progress bar
                self._clear_progress_bar()

                # Reactivate back button
                self.view.current_page.back_button.config(state='normal')

                # Replace Stop button with Exit button
                self.view.current_page.stop_button.pack_forget()
                self.view.current_page.exit_button.pack(side='left')

                return
        # If the queue is empty, continually check
        except queue.Empty:
            self.view.after(100, self.process_runtime_updates)

    def _clear_progress_bar(self) -> None:
        self.view.current_page.progress_bar.stop()
        self.view.current_page.progress_bar['mode'] = 'determinate'
        self.view.current_page.progress_bar['value'] = 0

    def _update_progress_bar_indeterminate(self) -> None:
        self.view.current_page.progress_bar['mode'] = 'indeterminate'
        if self.model.run_thread.collector.num_queries:
            self.view.current_page.progress_bar.start()
        else:
            self._clear_progress_bar()

    def _update_progress_bar_determinate(self) -> None:
        self.view.current_page.progress_bar.stop()
        self.view.current_page.progress_determinate_num['text'] = \
            f'({self.model.run_thread.collector.queries_completed}/' \
            f'{self.model.run_thread.collector.num_queries})'
        self.view.current_page.progress_bar['mode'] = 'determinate'
        self.view.current_page.progress_bar['value'] = \
            (self.model.run_thread.collector.queries_completed /
             self.model.run_thread.collector.num_queries * 100) + 1

    def _update_progress_bar(self) -> None:
        """Update status of progress bar based on collector status."""
        self.view.current_page.progress_label['text'] = \
            self.model.run_thread.collector.current_query_ref

        if isinstance(self.model.run_thread.collector.num_queries, bool):
            self._update_progress_bar_indeterminate()
        else:
            self._update_progress_bar_determinate()

    def _update_runtime_output(
            self,
            update: str,
            loc: _tkinter.Tcl_Obj | str | float | tk.Misc = 'end',
            newline: bool = True
    ) -> None:
        """Push Collector status to PyCurator UI window.

        Parameters
        ----------
        update : str
            Status to output.
        loc _tkinter.Tcl_Obj or str or float or tk.Misc
                optional (default='end')
            Position to place update at.
        newline : bool, optional (default=True)
            Flag for adding a newline to output. Default to True in
            order for updates to appear one after the other for easy
            user digest.
        """

        if newline:
            update = f'{update} \n'

        # Add new output
        self.view.current_page.runtime_output.config(state='normal')
        self.view.current_page.runtime_output.insert(loc, update)
        self.view.current_page.runtime_output.config(state='disabled')

        self.view.current_page.tkraise()



import tkinter as tk
import tkinter.ttk as ttk
from collections.abc import Iterable
from typing import Any, NoReturn

import pycurator.collectors
from pycurator import utils
from .base import ViewPage


class SelectionPage(ViewPage):
    """Selection Page of the PyCurator View component.

    Parameters
    ----------
    *args : tuple, optional
        Positional arguments for ViewPage constructor, which further
        passes to ttk.Frame.
    **kwargs : dict, optional
        Keyword arguments for ViewPage constructor, which further
        passes to ttk.Frame.

    Attributes
    ----------
    current_repository : str

    See Also
    --------
    pycurator.gui.base.ViewPage
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.current_repository = None

        # Create global selection frame/widgets
        self.selector_frame = ttk.Frame(master=self)
        self.repo_listbox = tk.Listbox(master=self.selector_frame)
        self.repo_listbox.bind(
            '<<ListboxSelect>>',
            self.display_repo_params
        )

        for repo_name in pycurator.collectors.available_repos:
            self.repo_listbox.insert(tk.END, repo_name)

        self.param_frame = ttk.Frame(master=self)

        self.req_var = tk.StringVar()
        self.req_label = ttk.Label(
            master=self.param_frame,
            foreground='#FF0000',
            textvariable=self.req_var
        )

    @ViewPage.no_overwrite
    def show(self, *args: Any) -> None:
        """Display collector selection widgets."""
        selection_text = ttk.Label(
            master=self.selector_frame,
            text='Select Repository:',
            anchor='center',
            font='helvetica 14 bold'
        )

        # Align widgets
        selection_text.grid(
            sticky='n',
            pady=(0, 10)
        )
        self.repo_listbox.grid(sticky='n')
        self.selector_frame.grid(row=0, column=0, sticky='n', padx=10, pady=5)
        self.param_frame.grid(row=0, column=1, sticky='n', padx=10, pady=5)
        self.grid(row=0, column=0, sticky='nsew')
        self.tkraise()

    def _set_model_from_name(self, repo_name) -> None:
        """Update Model element via Controller."""
        repo_class = pycurator.collectors.available_repos[repo_name]
        self.controller.set_model(repo_class, repo_name)

    def _clear_frame(self, frame: tk.Frame):
        """Delete widgets present on the provided frame."""
        for widget in frame.winfo_children():
            widget.destroy()

    def _get_selected_repo(self):
        """Get repository name from listbox selection."""
        selection_tuple = self.repo_listbox.curselection()
        if selection_tuple:
            return self.repo_listbox.get(selection_tuple[0])
        else:
            return None

    def display_repo_params(self, *args: Any) -> None:
        """Create and display Frame with repo-specific query parameters."""
        repo_selection = self._get_selected_repo()

        # No selection: nothing to show. No new selection: do not update
        if not repo_selection or repo_selection == self.current_repository:
            return

        self.current_repository = repo_selection
        self._clear_frame(self.param_frame)
        self._set_model_from_name(repo_selection)
        self.repo_listbox.selection_clear(0, tk.END)

        label = ttk.Label(
            master=self.param_frame,
            text='Parameter Selection:',
            font='helvetica 14 bold'
        )
        label.grid(sticky='nw')

        # Initialize run button
        self.next_page_button = ttk.Button(
            master=self.param_frame,
            text=f'Run {self.controller.model.collector_name}',
            command=self.controller.parse_run_parameters
        )

        # Get save information
        self.controller.add_run_parameter('save_type', tk.StringVar())
        save_dir_selection = utils.widget_label_frame(
            frame_master=self.param_frame,
            label_text='Save Directory:',
            widget_cls=ttk.Button,
            text='Select Directory',
            command=lambda: utils.select_from_files(
                root=self,
                selection_type='save_dir'
            )
        )
        save_frame = tk.Frame(master=self.param_frame)
        save_label = ttk.Label(master=save_frame, text='File Type:')
        save_type_menu = ttk.Combobox(
            master=save_frame,
            textvariable=self.controller.get_run_parameter('save_type'),
            values=[
                output_format for output_format in utils.save_options.keys()
            ],
            state='readonly'
        )
        save_label.grid(row=0, column=0)
        save_dir_selection.grid(row=0, columnspan=2, sticky='w')
        save_type_menu.grid(row=0, column=1)
        save_frame.grid()

        # Get credentials
        if self.controller.model.collector_class.accepts_user_credentials():
            user_credential_selection = utils.widget_label_frame(
                frame_master=self.param_frame,
                label_text='Credentials:',
                widget_cls=ttk.Button,
                text='Select File',
                command=lambda: utils.select_from_files(
                    root=self,
                    selection_type='credentials',
                    filetypes=[('JSON Files', '*.json')]
                )
            )
            user_credential_selection.grid(columnspan=2, sticky='w')

        # Get search terms, if needed
        if self.controller.model.requirements.get('search_terms'):
            search_term_frame = ttk.Frame(master=self.param_frame)
            self.controller.add_run_parameter('search_terms', tk.StringVar())

            search_term_label = ttk.Label(
                master=search_term_frame,
                text='Search Term(s):'
            )
            search_term_entry = ttk.Entry(
                master=search_term_frame,
                textvariable=self.controller.get_run_parameter('search_terms')
            )
            search_term_req = ttk.Label(
                master=search_term_frame,
                foreground='#FF0000',
                text='*'
            )

            search_term_req.grid(row=0, column=0, sticky='w')
            search_term_label.grid(row=0, column=1)
            search_term_entry.grid(row=0, column=2)
            search_term_frame.grid(columnspan=2)

        # Get search types, if needed
        if self.controller.model.requirements.get('search_types'):
            search_type_options = self.controller \
                .model \
                .collector_class. \
                search_type_options
            search_type_outer_frame = ttk.Frame(master=self.param_frame)
            search_type_inner_frame = ttk.Frame(master=search_type_outer_frame)
            search_type_label = ttk.Label(
                master=search_type_outer_frame,
                text='Search Type(s):'
            )
            search_type_req = ttk.Label(
                master=search_type_outer_frame,
                foreground='#FF0000',
                text='*'
            )

            self.controller.add_run_parameter(
                param='search_types',
                value={
                    search_type: tk.IntVar()
                    for search_type in search_type_options
                }
            )

            for search_type in search_type_options:
                search_type_button = ttk.Checkbutton(
                    master=search_type_inner_frame,
                    text=search_type.title(),
                    variable=self.controller.get_run_parameter(
                        'search_types'
                    )[search_type]
                )
                search_type_button.grid(sticky='w')

            search_type_req.grid(row=0, column=0, sticky='w')
            search_type_label.grid(row=0, column=1, sticky='n')
            search_type_inner_frame.grid(column=1, sticky='w')
            search_type_outer_frame.grid(sticky='w')

        # Run button
        self.next_page_button.grid()

    def alert_missing_reqs(self, missing_requirements: list[str]) -> None:
        """Display unfulfilled requirements when attempting to run."""
        try:
            self.req_label.grid_forget()
        except tk.TclError:
            pass
        finally:
            req_list = '\n'.join(
                [
                    f'\u2022 {" ".join(req.split("_")).title()}'
                    for req in missing_requirements
                ]
            )
            self.req_var.set(
                f'Missing Requirements:\n{req_list}'
            )
            self.req_label = ttk.Label(
                master=self.param_frame,
                foreground='#FF0000',
                textvariable=self.req_var
            )
        self.req_label.grid(sticky='nsew')

    def _toggle_button_state(
            self,
            toggle_vars: Iterable[tk.Variable],
            btn: tk.Button
    ) -> None:
        """Modify button state based on values of provided variables.

        Parameters
        ----------
        toggle_vars : tk.Variable or iterable of tk.Variable
            Variables on which value state is used to update
            button state.
        btn : tk.Button
            Widget to update state of.

        Raises
        ------
        TypeError
            Incorrect variable types provided to toggle_vars.
        """
        if not btn:
            return

        # Validate input
        if isinstance(toggle_vars, tk.Variable):
            toggle_vars = [toggle_vars]
        if not all([hasattr(var, 'get') for var in toggle_vars]):
            raise TypeError(
                'All entries of toggle_vars must implement "get" method.'
            )

        # Change button state if any of the passed variables are active
        if any([var.get() for var in toggle_vars]):
            btn.config(state=tk.NORMAL)
        else:
            btn.config(state=tk.DISABLED)

    def reset_frame(self) -> NoReturn:
        pass



import itertools
import json
import os
import queue
import sys
import time
from abc import ABC, abstractmethod
from collections.abc import (
    Iterable,
    Collection,
    Callable,
    Generator,
    Hashable,
    Sequence
)
from typing import Any, AnyStr, NoReturn, Optional, ParamSpec, TypeVar, Union

import pandas as pd
import requests

from pycurator import utils
from pycurator._typing import (
    JSONDict,
    SearchTerm,
    SearchType,
    SearchTuple,
    TermResultDict,
    TermTypeResultDict,
    QueryResultDict
)

T = TypeVar('T')
P = ParamSpec('P')


class BaseCollector(ABC):
    """Generic abstract base for data-collection classes.

    Parameters
    ----------
    repository_name : str
        Name of the repository being collected from. Used for providing
        updates to user, loading credentials, and saving output results.

    Attributes
    ----------
    continue_running : bool
        Flag when the collector completes a run. Used for pushing
        updates from the object's status_queue.
    current_query_ref : str or None (default=None)
        Representation of the current state of the collector run.
    num_queries : int or bool or None (default=None)
        Number of queries for a given run.
        If there is no fixed number, such as for paginated queries, the
        variable is True.
    queries_completed : int or None (default=None)
    status_queue : queue.Queue of str
        FIFO collection of the object's status.

    See Also
    --------
    queue.Queue : Queue data structure
    BaseAPICollector : Derived Class for API queries.
    """

    def __init__(self, repository_name: str) -> None:
        self.repository_name = repository_name
        self.continue_running = True

        self.status_queue = queue.Queue()
        self.num_queries = None
        self.queries_completed = None
        self.current_query_ref = None

    def _pb_determinate(self, coll: Collection[T]) -> Generator[T, None, None]:
        """Generator for iterating data and updating progress bar.

        Parameters
        ----------
        coll : iterable

        Yields
        ------
        next object of coll

        Raises
        ------
        TypeError
            coll parameter is not iterable.
        """

        if not hasattr(coll, '__iter__'):
            raise TypeError('Parameter "coll" must be iterable.')

        # Initialize tracking vars
        if not self.num_queries:
            self.num_queries = len(coll)
            self.queries_completed = 0

        # Yield next item in coll and update tracking vars
        for item in coll:
            self.current_query_ref = str(item)
            yield item
            self.queries_completed += 1

        # Reset tracking vars
        self.num_queries = None
        self.queries_completed = None
        self.current_query_ref = None

    def _save_results(
            self,
            save_dir: str,
            final_dict: QueryResultDict,
            output_format: str
    ) -> None:
        self.status_queue.put(
            f'Saving output to "{save_dir}".'
        )
        utils.save_results(
            results=final_dict,
            data_dir=save_dir,
            output_format=output_format
        )
        self.status_queue.put('Save complete.')

    @abstractmethod
    def run(self) -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "run()".'
        )

    @staticmethod
    def _pb_indeterminate(
            indeterminate_query_func: Callable[P, T]
    ) -> Callable[P, T]:
        """Progress bar wrapper for indeterminate-length queries."""

        def update_pb(
                self,
                *args: P.args,
                **kwargs: P.kwargs
        ) -> Iterable[Any]:
            self.num_queries = True
            results = indeterminate_query_func(self, *args, **kwargs)
            self.num_queries = False
            return results

        return update_pb

    def request_execution(self) -> None:
        """Raise flag to stop output."""
        self.continue_running = False

    def terminate(self) -> NoReturn:
        """Handle program execution."""
        self.status_queue.put('Requesting program termination.')
        sys.exit()

    @staticmethod
    @abstractmethod
    def accepts_user_credentials() -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "accepts_user_credentials()".'
        )

    def _print_progress(self, page: str) -> None:
        """Update queue with current page being searched."""
        self.status_queue.put(f'Searching page {page}')

    def _update_query_ref(self, **kwargs: Any) -> None:
        """Combine keywords and update base.current_query_ref."""
        self.current_query_ref = kwargs


class BaseAPICollector(BaseCollector):
    """Base for collection classes utilizing external API.

    This base inherits from BaseCollector, which provides general
    parameters for tracking collection progress.

    Parameters
    ----------
    repository_name : str
        Name of the repository being collected from. Used for providing
        updates to user, loading credentials, and saving output results.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.

    Attributes
    ----------
    credentials : str
        Refer to credentials parameter.

    See Also
    --------
    BaseTermCollector
    BaseTermTypeCollector
    BaseTypeCollector
    """

    def __init__(
            self,
            repository_name: str,
            credentials: Optional[str] = None
    ) -> None:
        super().__init__(repository_name=repository_name)

        if credentials:
            self.credentials = self.load_credentials(
                credential_filepath=credentials
            )

    def load_credentials(self, credential_filepath: str) -> Union[str, None]:
        """Load the credential file from the given filepath.

        Parameters
        ----------
        credential_filepath : str or path-like object

        Returns
        -------
        credentials : str or None

        Raises
        ------
        ValueError
            If credential_filepath is not of type str.
        FileNotFoundError
            Credentials file does not exist.

        See Also
        --------
        os.path : Module for functions on pathnames.
        """

        if not isinstance(credential_filepath, str):
            raise TypeError(
                (f'Credential value must be of type str, '
                    f'not \'{type(credential_filepath)}\'.')
            )

        # Try to load credentials from file
        if os.path.exists(credential_filepath):
            with open(credential_filepath) as credential_file:
                credential_data = json.load(credential_file)

                credentials = credential_data.get(self.repository_name)

                if not credentials:
                    self.status_queue.put(
                        'No credentials found, attempting unauthorized run.'
                    )
                return credentials
        else:
            raise FileNotFoundError(f'{credential_filepath} does not exist.')

    @staticmethod
    def accepts_user_credentials() -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "accepts_user_credentials()".'
        )

    @staticmethod
    def _all_empty(data_dict: QueryResultDict) -> bool:
        """Check if all DataFrames are empty.

        Parameters
        ----------
        data_dict : dict from search params to pandas.DataFrame.

        Returns
        -------
        all_empty : bool
        """

        return all([df is None or df.empty for df in data_dict.values()])

    def run(self, **kwargs: Any) -> None:
        """Queries all data from the implemented API.

        Parameters
        ----------
        **kwargs : dict, optional
            Can temporarily overwrite base attributes.
            Allows users to specify variable save parameters.

        Notes
        -----
        In the following order, this function calls:
            get_all_search_outputs
            get_all_metadata (if applicable)
            merge_search_and_metadata_dicts (if applicable)
        """

        self.status_queue.put(f'Running {self.repository_name.title()}...')

        # Set save parameters
        save_dir = kwargs.pop('save_dir', None)
        save_type = kwargs.pop('save_type', None)

        try:
            # Get search_output
            search_dict = self.get_all_search_outputs(**kwargs)

            # Set merge parameters
            merge_on = vars(self).get('merge_on')
            merge_right_on = vars(self).get('merge_right_on')
            merge_left_on = vars(self).get('merge_left_on')

            # Try to get metadata (if available)
            try:
                metadata_dict = self.get_all_metadata(search_dict=search_dict)
                merged_dict = self.merge_search_and_metadata_dicts(
                    search_dict=search_dict,
                    metadata_dict=metadata_dict,
                    on=merge_on,
                    left_on=merge_left_on,
                    right_on=merge_right_on
                )
                final_dict = merged_dict
            except TypeError:  # Function with incorrect parameters
                final_dict = search_dict
        except Exception as e:
            self.status_queue.put(f'An unexpected error has occurred: \n{e}')
            self.continue_running = False
            return

        # Handle saving if output exists
        if not self._all_empty(final_dict):
            if save_dir and save_type:
                self._save_results(
                    save_dir=save_dir,
                    final_dict=final_dict,
                    output_format=save_type
                )
        else:
            self.status_queue.put('No results found, nothing to save.')

        self.status_queue.put(f'{self.repository_name.title()} run complete.')
        self.continue_running = False

    def get_all_search_outputs(self, **kwargs: Any) -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "get_all_search_outputs()".'
        )

    def get_all_metadata(self, **kwargs: Any) -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "get_all_metadata()".'
        )

    def get_request_output_and_update_query_ref(
            self,
            url: AnyStr,
            params: Optional[Any] = None,
            headers: Optional[Any] = None,
            **ref_kwargs: Any
    ) -> tuple[requests.Response, JSONDict]:
        """Return request output and update base.current_query_ref.

        Parameters
        ----------
        url : str
        params : dict, optional (default=None)
        headers : dict, optional (default=None)
        **ref_kwargs : dict, optional

        Returns
        -------
        base.get_request_output(url, params, headers)

        See Also
        --------
        _update_query_ref
        get_request_output
        """

        self._update_query_ref(**ref_kwargs)
        return self.get_request_output(url=url, params=params, headers=headers)

    def get_request_output(
            self,
            url: AnyStr,
            params: Optional[Any] = None,
            headers: Optional[Any] = None
    ) -> tuple[requests.Response, JSONDict]:
        """Return Response and JSON from requests.get().

        Parameters
        ----------
        url : str
        params : dict or list of tuples or bytes, optional
                (default=None)
            Dictionary, list of types or bytes to send in the query
            string for the Request.
        headers : dict, optional (default=None)
            Dictionary of headers to send with the Request.

        Returns
        -------
        r : requests.Response
        output : dict
            JSON-encoded content of a response.

        Raises
        ------
        RuntimeError
            Occurs when a query results in an unparsable response.
            Outputs the parameters provided to the query along with the
            response status code for further troubleshooting.

        See Also
        --------
        requests.get : Sends a GET request.
        """

        # If user has requested termination, handle cleanup instead of querying
        # additional results
        if not self.continue_running:
            self.terminate()

        r = requests.get(url=url, params=params, headers=headers)
        try:
            output = r.json()
        except json.JSONDecodeError:
            # 429: Rate limiting (wait and then try the request again)
            if r.status_code == 429:
                self.status_queue.put('Rate limit hit, waiting for request...')

                # Wait until we can make another request
                reset_time = int(r.headers['RateLimit-Reset'])
                current_time = int(time.time())
                time.sleep(reset_time - current_time)

                r, output = self.get_request_output(
                    url=url,
                    params=params,
                    headers=headers
                )
            else:
                raise RuntimeError(
                    (f'Query to {url} with {params} params and {headers}'
                     f' headers fails unexpectedly with status'
                     f' code {r.status_code} and full output {vars(r)}')
                )

        return r, output

    def merge_search_and_metadata_dicts(
            self,
            search_dict: QueryResultDict,
            metadata_dict: dict,
            on: Optional[Union[Hashable, Sequence[Hashable]]] = None,
            left_on: Optional[Union[Hashable, Sequence[Hashable]]] = None,
            right_on: Optional[Union[Hashable, Sequence[Hashable]]] = None,
            **kwargs: Any
    ) -> QueryResultDict:
        """Merges together search and metadata DataFrames by 'on' key.

        For multiple DataFrames containing similar search references,
        combines into one DataFrame. Search and Metadata DataFrames are
        merged across their respective dictionaries via common keys.
        For Search DataFrames with no matching Metadata, the Search
        DataFrame is added as-is.

        Parameters
        ----------
        search_dict : dict of pandas.DataFrame
            Dictionary of search output results.
        metadata_dict : dict of pandas.DataFrame
            Dictionary of metadata results.
        on : str or list-like, optional (default=None)
            Column name(s) to merge the two dicts on.
        left_on : str or list-like, optional (default=None)
            Column name(s) to merge the left dict on.
        right_on : str or list-like, optional (default=None)
            Column name(s) to merge the right dict on.
        **kwargs : dict, optional
            Placeholder argument to allow inheritance overloading.

        Returns
        -------
        df_dict : dict of pandas.DataFrame
            Dict containing all the merged search/metadata DataFrames
            or singleton search DataFrames.

        Raises
        ------
        TypeError
            search_dict or metadata_dict are not instances of dict.
        ValueError
            search_dict or metadata_dict contain entries that are not of
            type pandas.DataFrame.

        See Also
        --------
        pandas.merge
        """

        if not isinstance(search_dict, dict):
            raise TypeError(
                ('search_dict must be of type dict, not'
                 f'\'{type(search_dict)}\'.')
            )
        if not isinstance(metadata_dict, dict):
            raise TypeError(
                ('metadata_dict must be of type dict, not'
                 f' \'{type(metadata_dict)}\'.')
            )

        if not all(
                [
                    isinstance(df, pd.DataFrame) or df is None
                    for df in search_dict.values()
                ]
        ):
            raise ValueError(
                'All search_dict entries must be of type pandas.DataFrame.'
            )
        if not all(
                [
                    isinstance(df, pd.DataFrame) or df is None
                    for df in metadata_dict.values()
                ]
        ):
            raise ValueError(
                'All metadata_dict entries must be of type pandas.DataFrame.'
            )

        df_dict = dict()
        for query_key in search_dict.keys():
            search_df = search_dict[query_key]

            # If the search DataFrame has matching metadata, merge
            if query_key in metadata_dict:
                metadata_df = metadata_dict[query_key]
                df_all = pd.merge(
                    left=search_df.convert_dtypes(),
                    right=metadata_df.convert_dtypes(),
                    on=on,
                    left_on=left_on,
                    right_on=right_on,
                    how='outer',
                    suffixes=('_search', '_metadata')
                )
            # If no metadata, just add the search_df
            else:
                df_all = search_df

            df_dict[query_key] = df_all

        return df_dict


class TermQueryMixin:
    """Mixin for API collection classes that utilize search terms.

    Attributes
    ----------
    search_terms : list of str

    See Also
    --------
    BaseTermCollector
    BaseTermTypeCollector
    """

    @property
    def search_terms(self) -> Collection[SearchTerm]:
        return self._search_terms

    @staticmethod
    def validate_search_term(f: Callable[P, T]) -> Callable[P, T]:
        """Decorator for validating search term object type."""
        def inner(self, *args, **kwargs):
            args, kwargs = utils.validate_from_arguments(
                base=self,
                func=f,
                args=args,
                kwargs=kwargs,
                param='search_term'
            )
            return f(self, *args, **kwargs)
        return inner

    @search_terms.setter
    def search_terms(self, search_terms: Collection[SearchTerm]) -> None:
        if isinstance(search_terms, str):
            search_terms = [search_terms]
        if not all([isinstance(term, str) for term in search_terms]):
            raise TypeError('All search terms must be of type str.')
        self._search_terms = search_terms

    def _validate(self, search_term: SearchTerm) -> SearchTerm:
        """Validate type of search_term."""
        if not isinstance(search_term, str):
            raise TypeError(
                'search_term must be of type str, not'
                f' \'{type(search_term)}\'.'
            )
        else:
            return search_term


class BaseTermCollector(TermQueryMixin, BaseAPICollector):
    """Base for API collection classes that utilize search terms.

    This base inherits from BaseAPICollector, which provides credential
    info, as well as general parameters for tracking collection
    progress, inherited from BaseCollector.

    Parameters
    ----------
    repository_name : str
        Name of the repository being collected from. Used for providing
        updates to user, loading credentials, and saving output results.
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms() or
        passed in directly to search functions to override set parameter.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.

    Attributes
    ----------
    search_terms : list of str

    See Also
    --------
    pycurator.collectors.term_collectors
    """

    def __init__(
            self,
            repository_name: str,
            search_terms: Optional[Collection[SearchTerm]] = None,
            credentials: Optional[str] = None
    ) -> None:
        super().__init__(
            repository_name=repository_name,
            credentials=credentials
        )

        self.search_terms = search_terms

    def get_all_search_outputs(
            self,
            **kwargs: Any
    ) -> TermResultDict:
        """Queries the API for each search term.

        Parameters
        ----------
        **kwargs : dict, optional
            Can temporarily overwrite self search_terms.

        Returns
        -------
        search_dict : dict of pandas.DataFrame
            Stores the results of each call to
            get_individual_search_output in the form
            search_dict[{search_term}] = df.
        """

        # Set method variables if different than default values
        search_terms = kwargs.get('search_terms', self.search_terms)

        search_dict = dict()

        for search_term in search_terms:
            self.status_queue.put(f'Searching {search_term}.')
            search_dict[search_term] = self.get_individual_search_output(
                search_term=search_term
            )
            self.status_queue.put('Search completed.')

        return search_dict

    @abstractmethod
    def get_individual_search_output(self, search_term: SearchTerm) -> None:
        pass

    def get_all_metadata(
            self,
            object_path_dict: dict[SearchTerm, pd.DataFrame]
    ) -> TermResultDict:
        """Retrieves all metadata related to the provided DataFrames.

        Parameters
        ----------
        object_path_dict : dict
            Dictionary of the form {query: object_paths} for path lists.

        Returns
        -------
        metadata_dict : dict of {SearchTerm: pd.DataFrame}
        """

        metadata_dict = dict()

        for query, object_paths in object_path_dict.items():
            self.status_queue.put(f'Querying {query} metadata.')
            metadata_dict[query] = self.get_query_metadata(
                object_paths=object_paths
            )
            self.status_queue.put('Metadata query complete.')

        return metadata_dict

    def get_query_metadata(
            self,
            object_paths: Iterable[Any]
    ) -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "get_query_metadata()".'
        )


class TypeQueryMixin:
    """Mixin for API collection classes that utilize search types.

    Attributes
    ----------
    search_types : list of str

    See Also
    --------
    BaseTermTypeCollector
    BaseTypeCollector
    """

    @property
    def search_types(self) -> tuple[SearchType]:
        """Getter for search_types."""
        return self._search_types

    @search_types.setter
    def search_types(self, search_types: tuple[SearchType]) -> None:
        """Set search_types if all are allowed by current Collector."""
        if not all(
                [
                    search_type in self.search_type_options
                    for search_type in search_types
                ]
        ):
            raise ValueError(
                f'Only {self.search_type_options} search types are valid.'
            )
        self._search_types = search_types

    @staticmethod
    def validate_search_type(f: Callable[P, T]) -> Callable[P, T]:
        """Decorator for validating search term object type."""
        def inner(self, *args, **kwargs):
            args, kwargs = utils.validate_from_arguments(
                base=self,
                func=f,
                args=args,
                kwargs=kwargs,
                param='search_type'
            )
            return f(self, *args, **kwargs)
        return inner

    def _validate(self, search_type: SearchType) -> SearchType:
        if search_type not in self.search_type_options:
            raise ValueError(
                f'Can only search by {self.search_type_options}.'
            )
        else:
            return search_type

    @classmethod
    @abstractmethod
    def search_type_options(cls) -> NoReturn:
        """Return the valid search type options for a given repository."""
        raise NotImplementedError(
            'Subclass must override "search_type_options()".'
        )


class BaseTermTypeCollector(
    TermQueryMixin,
    TypeQueryMixin,
    BaseAPICollector
):
    """Base for API collection classes that utilize search terms and types.

    This base inherits from BaseAPICollector, which provides credential
    info, as well as general parameters for tracking collection
    progress, inherited from BaseCollector.

    Parameters
    ----------
    repository_name : str
        Name of the repository being collected from. Used for providing
        updates to user, loading credentials, and saving output results.
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms() or
        passed in directly to search functions to override set parameter.
    search_types : list-like, optional (default=None)
        Data types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions to override set
        parameter.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.

    Attributes
    ----------
    search_terms : list of str
    search_types : list of str

    See Also
    --------
    pycurator.collectors.term_type_collectors
    """

    def __init__(
            self,
            repository_name: str,
            search_terms: Optional[Collection[SearchTerm]] = None,
            search_types: Optional[Collection[SearchType]] = None,
            credentials: Optional[str] = None
    ) -> None:
        super().__init__(
            repository_name=repository_name,
            credentials=credentials
        )

        self.search_terms = search_terms
        self.search_types = search_types

    @staticmethod
    def validate_term_and_type(f: Callable[P, T]) -> Callable[P, T]:
        """Helper for wrapping function in both term/type validators."""
        @BaseTermTypeCollector.validate_search_term
        @BaseTermTypeCollector.validate_search_type
        def inner(self, *args, **kwargs):
            return f(self, *args, **kwargs)

        return inner

    def get_all_search_outputs(
            self,
            **kwargs: Any
    ) -> TermTypeResultDict:
        """Queries the API for each search term/type combination.

        Parameters
        ----------
        **kwargs : dict, optional
            Can temporarily overwrite self search_terms and
            search_types.

        Returns
        -------
        search_dict : dict of pandas.DataFrame
            Stores the results of each call to
            get_individual_search_output in the form
            search_dict[(search_term, search_type)] = df.
        """

        # Set method variables if different than default values.
        search_terms = kwargs.get('search_terms', self.search_terms)
        search_types = kwargs.get('search_types', self.search_types)

        search_dict = dict()

        for search_term, search_type in itertools.product(
                search_terms, search_types
        ):
            self.status_queue.put(f'Searching {search_term} {search_type}.')
            search_dict[(search_term, search_type)] = \
                self.get_individual_search_output(
                    search_term=search_term,
                    search_type=search_type
            )
            self.status_queue.put('Search completed.')

        return search_dict

    @abstractmethod
    def get_individual_search_output(
            self,
            search_term: SearchTerm,
            search_type: SearchType
    ) -> None:
        pass

    def get_all_metadata(
            self,
            object_path_dict: dict[SearchTuple, Collection[str]]
    ) -> TermTypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        object_path_dict : dict
            Dictionary of the form {query: object_paths} for path lists.

        Returns
        -------
        metadata_dict : dict of {(SearchTerm, SearchType): pd.DataFrame}
        """

        metadata_dict = dict()

        for query, object_paths in object_path_dict.items():
            search_term, search_type = query
            self.status_queue.put(
                f'Querying {search_term} {search_type} metadata.'
            )

            metadata_dict[query] = self.get_query_metadata(
                object_paths=object_paths
            )
            self.status_queue.put('Metadata query complete.')

        return metadata_dict

    def get_query_metadata(self, object_paths: Collection[str]) -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "get_query_metadata()".'
        )


class BaseTypeCollector(TypeQueryMixin, BaseAPICollector):
    """Base for API collection classes that utilize search types.

    This base inherits from BaseAPICollector, which provides credential
    info, as well as general parameters for tracking collection
    progress, inherited from BaseCollector.

    Parameters
    ----------
    repository_name : str
        Name of the repository being collected from. Used for providing
        updates to user, loading credentials, and saving output results.
    search_types : list-like, optional (default=None)
        types to search over. Can be (re)set via set_search_types() or
        passed in directly to search functions to override set parameter.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.

    Attributes
    ----------
    search_types : list of str

    See Also
    --------
    pycurator.collectors.type_collectors
    """

    def __init__(
            self,
            repository_name: str,
            search_types: Optional[Collection[SearchType]] = None,
            credentials: Optional[str] = None
    ) -> None:
        super().__init__(
            repository_name=repository_name,
            credentials=credentials
        )

        self.search_types = search_types

    @property
    def search_types(self) -> Collection[SearchType]:
        return self._search_types

    @search_types.setter
    def search_types(self, search_types: Collection[SearchType]) -> None:
        if not search_types:
            return
        if not all(
            [
                search_type in self.search_type_options
                for search_type in search_types
            ]
        ):
            raise ValueError(
                f'Only {self.search_type_options} search types are valid.'
            )
        self._search_types = search_types

    @classmethod
    @abstractmethod
    def search_type_options(cls) -> None:
        """Return the valid search type options for a given repository."""
        pass

    def get_all_search_outputs(
            self,
            **kwargs: Any
    ) -> TermResultDict:
        """Queries the API for each search type.

        Parameters
        ----------
        **kwargs : dict, optional
            Can temporarily overwrite self search_types.

        Returns
        -------
        search_dict : dict of pandas.DataFrame
            Stores the results of each call to
            get_individual_search_output in the form
            search_output_dict[{search_type}] = df.
        """

        search_types = kwargs.get('search_types', self.search_types)

        search_dict = dict()

        for search_type in search_types:
            self.status_queue.put(f'Searching {search_type}.')
            search_dict[search_type] = self.get_individual_search_output(
                search_type=search_type
            )
            self.status_queue.put(f'{search_type} search completed.')

        return search_dict

    @abstractmethod
    def get_individual_search_output(self, search_type: SearchType) -> None:
        pass

    def get_query_metadata(
            self,
            object_paths: Collection[str],
            search_type: SearchType,
    ) -> NoReturn:
        raise NotImplementedError(
            'Subclass must override "get_query_metadata()".'
        )



from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import ParamSpec, Type, TypeVar


from pycurator import gui

P = ParamSpec('P')
T = TypeVar('T')


def widget_label_frame(
        frame_master: tk.Misc,
        label_text: float | str,
        widget_cls: Type[tk.Widget],
        **widget_kwargs: P.kwargs
) -> ttk.Frame:
    """Create frame containing vertically-aligned widget and label.

    Parameters
    ----------
    frame_master : tk.Misc
    label_text : float or str
    widget_cls : tk.Widget
    **widget_kwargs

    Returns
    -------
    _frame : ttk.Frame

    Examples
    --------
    >>> root = tk.Tk()
    >>> widget_label_frame(
        frame_master=root,
        label_text="Don't click this button:",
        widget_cls=ttk.Button,
        widget_text='Click me!',
        widget_command=sys.exit
    )
    """

    _frame = ttk.Frame(master=frame_master)
    _label = ttk.Label(master=_frame, text=label_text)
    _widget = widget_cls(
        master=_frame,
        **widget_kwargs
    )

    _label.grid(row=0, column=0)
    _widget.grid(row=0, column=1)
    return _frame


def select_from_files(
        root: gui.ViewPage,
        selection_type: str,
        filetypes: list[tuple[str, str]] = (('All File Types', '*.*'),)
) -> None:
    """Allows user to select local file/directory.

    Parameters
    ----------
    root : Tkinter.Frame
        Derived Frame class that contains UI.
        If choosing file, frame_master must contain "files" dict attribute to
        hold the selected file.
    selection_type : str
        Type of selection to be made.
        Examples include "credentials", "directory", "css_paths" etc.
    filetypes : list of two-tuples of str,
            optional (default = [('All File Types', '*.*')])
        Allows users to input filetype restrictions.
        Each tuple should be of the form:
            ('File Type', '{file_name}.{file_format}')

    Raises
    ------
    NotImplementedError
        Occurs when called on a frame_master that does not contain a files
        attribute.

    Examples
    --------
    >>> f = tk.Frame()
    >>> select_from_files(
    ...     root=f,
    ...     selection_type='dir'
    ... )
    Traceback (most recent call last):
        ...
    NotImplementedError: 'root' must include add_run_parameter() method.'

    >>> from pycurator.gui import ViewPage
    >>> view = ViewPage()
    >>> select_from_files(
    ...     root=view,
    ...     selection_type='dir',
    ... )
    """

    from tkinter import filedialog as fd

    if 'dir' in selection_type:
        selection = fd.askdirectory(
            title='Choose Directory',
            initialdir='.'
        )
    else:
        selection = fd.askopenfilename(
            title='Choose File',
            initialdir='.',
            filetypes=filetypes
        )

    try:
        root.controller.add_run_parameter(selection_type, selection)
    except AttributeError:
        raise NotImplementedError(
            '\'root\' must include add_run_parameter() method.'
        )



from collections.abc import Collection
from typing import Optional, Union, NoReturn

import pandas as pd

from pycurator._typing import (
    SearchTerm,
    SearchType,
)
from pycurator.collectors import (
    BaseCollector,
    BaseTermTypeCollector,
)


class DataverseCollector(BaseTermTypeCollector):
    """Harvard Dataverse collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions.
    search_types : list-like, optional (default=None)
        Data types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions to override set
        parameter.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def get_query_metadata(self, object_paths: Collection[str]) -> NoReturn:
        pass

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        search_types: Optional[Collection[SearchType]] = None,
        credentials: Optional[str] = None,
    ) -> None:

        super().__init__(
            repository_name='dataverse',
            search_terms=search_terms,
            search_types=search_types
        )

        self.api_url = 'https://dataverse.harvard.edu/api'
        self.headers = dict()

        if credentials:
            self.credentials = self.load_credentials(
                credential_filepath=credentials
            )

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ('dataset', 'file')

    def load_credentials(self, credential_filepath: str) -> Union[str, None]:
        """Load the credentials given filepath or token.

        Parameters
        ----------
        credential_filepath : str,
            JSON filepath containing credentials in form
            {repository_name}: {key}.

        Returns
        -------
        credentials : str or None
        """

        credentials = super().load_credentials(
            credential_filepath=credential_filepath
        )
        self.headers['X-Dataverse-key'] = credentials
        return credentials

    @BaseCollector._pb_indeterminate
    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
            self,
            search_term: SearchTerm,
            search_type: SearchType
    ) -> pd.DataFrame:
        """Queries Dataverse API for the specified search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {'dataset', 'file'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        search_url = f'{self.api_url}/search'

        # Set search parameters
        start = 0
        page_size = 25
        search_df = pd.DataFrame()
        page_idx = 0

        search_params = {
            'q': search_term,
            'per_page': page_size,
            'start': start,
            'type': search_type
        }

        # Conduct initial search & extract results
        _, output = self.get_request_output_and_update_query_ref(
            url=search_url,
            params=search_params,
            headers=self.headers,
            search_term=search_term,
            search_type=search_type,
            page=page_idx
        )
        output = output['data']

        while output.get('items'):
            output = output['items']

            output_df = pd.DataFrame(output)
            output_df['page'] = (
                search_params['start'] // search_params['per_page'] + 1
            )

            search_df = pd.concat(
                [search_df, output_df]
            ).reset_index(drop=True)

            search_params['start'] += search_params['per_page']
            page_idx += 1

            _, output = self.get_request_output_and_update_query_ref(
                url=search_url,
                params=search_params,
                headers=self.headers,
                search_term=search_term,
                search_type=search_type,
                page=page_idx
            )
            output = output['data']

        return search_df



import logging
from collections.abc import Callable, Collection
from typing import Any, Optional, ParamSpec, TypeVar, Union


T = TypeVar('T')
P = ParamSpec('P')


def validate_from_arguments(
        base,
        param: str,
        func: Callable[P, T],
        args: Optional[tuple[Any, ...]] = None,
        kwargs: Optional[dict[str, Any]] = None,
):
    if not (args or kwargs):
        raise ValueError('args or kwargs must be provided to extract.')

    # Checking for param in kwargs
    if kwargs:
        if not isinstance(kwargs, dict):
            raise TypeError(
                'kwargs parameter must be of type \'dict\', not'
                f' \'{type(kwargs)}\'.'
            )
        if param in kwargs:
            kwargs[param] = base._validate(kwargs.get(param))
    else:
        kwargs = dict()

    # Checking for param in args
    if args:
        if not isinstance(args, tuple):
            raise TypeError(
                'args parameter must be of type \'set\', not' 
                ' \'{type(args)}\'.'
            )
        try:
            func_params = func.__code__.co_varnames
            idx = func_params.index(param)
            if 'self' in func_params:
                idx -= 1
            args = list(args)
            args[idx] = base._validate(args[idx])
            args = tuple(args)
        except (IndexError, ValueError):
            logging.debug(
                f'Attempting to validate \'{param}\' where one does not'
                ' appear to be present.'
            )
    else:
        args = set()

    return args, kwargs


def validate_metadata_parameters(
        object_paths: Union[str, Collection[str]]
) -> Collection[str]:
    """Ensures that the metadata object paths are of the proper form.

    Parameters
    ----------
    object_paths : str or Collection of str

    Returns
    -------
    object_paths : str or Collection of str
    Raises
    ------
    TypeError
        If object paths are not all str instance.
    """

    if isinstance(object_paths, str):
        object_paths = [object_paths]
    if not all([isinstance(path, str) for path in object_paths]):
        raise TypeError('All object paths must be of type str.')

    return object_paths


def validate_save_filename(filename: str) -> str:
    """Remove quotations from filename, replace spaces with underscore."""
    return filename.replace('"', '').replace("'", '').replace(' ', '_')



from __future__ import annotations

import threading
import tkinter as tk
import tkinter.ttk as ttk
from abc import abstractmethod
from collections.abc import Callable
from typing import Any, NoReturn

import pycurator.collectors
import pycurator.gui


class PyCuratorUI(tk.Tk):
    """UI for PyCurator Application.

    The UI is structured in a Model-View-Controller format. The View
    is separated into a Landing Page with license information, a
    Selection Page with repository options and related runtime
    parameters, and a Run Page providing real-time updates on the status
    of the data collection task.

    See Also
    --------
    pycurator.gui.MVC :
        Module containing the Model, View, and Controller classes.
    """

    def __init__(self):
        super().__init__()

        self.title('PyCurator')

        view = pycurator.gui.CuratorView(self)
        controller = pycurator.gui.CuratorController(view=view)

        self.bind('<Return>', controller.request_next_page)

        view.set_controller(controller)
        view.show()


class ThreadedRun(threading.Thread):
    """Wrapper for concrete Collector object to allow threading.

    Parameters
    ----------
    collector : Subclass of BaseCollector
    **kwargs : dict, optional
        Additional parameters to pass to the run function of the given
        collector object. Current integrated examples include the
        save_type of the output file and save_dir for storing the
        output file.

    See Also
    --------
    run :
        Start-to-finish pipeline for running Collectors. See specific
        repository Collector classes for more concrete details.
    """

    def __init__(
            self,
            collector: pycurator.collectors.BaseCollector,
            **kwargs: Any
    ) -> None:
        self.collector = collector
        super().__init__(target=self.collector.run, **kwargs)


class ViewPage(ttk.Frame):
    """Base for page in UI View object.

    Parameters
    ----------
    *args : tuple, optional
        Positional arguments to pass to ttk.Frame constructor.
    **kwargs : dict, optional
        Keyword arguments to pass to ttk.Frame constructor.

    Attributes
    ----------
    controller : MVC.CuratorController
        Default of None to allow UI instantiation of pages before
        Controller.
    is_initialized : bool
    next_page_button : None or ttk.Button
        Widget with action linking to the next sequential page of the UI.

    Methods
    -------
    set_controller
    no_overwrite
        Decorator for loading a page without overwriting the previously
        displayed values. Useful for reloading a page without destroying
        widgets or reconstructing an instance of the page.
    reset_frame
    attempt_raise
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ttk.Frame.__init__(self, *args, **kwargs)

        self.controller = None
        self.is_initialized = False
        self.next_page_button = None

    def set_controller(
            self,
            controller: pycurator.gui.MVC.CuratorController
    ) -> None:
        """Setter for Controller element of PyCurator UI."""
        self.controller = controller

    @staticmethod
    def no_overwrite(show_func: Callable[[Any], Any]) -> Callable[[Any], Any]:
        """Wrapper for page display without element overwrite."""
        def display(self) -> None:
            raised = self.attempt_raise()
            if not raised:
                show_func(self)
                self.is_initialized = True

        return display

    @abstractmethod
    def reset_frame(self) -> NoReturn:
        raise NotImplementedError

    def attempt_raise(self) -> bool:
        """Display page if initialized."""
        if self.is_initialized:
            try:
                self.reset_frame()
            except AttributeError:
                pass

            self.tkraise()
            return True
        return False



import os
from typing import Union

import pandas as pd

from .validating import validate_save_filename

save_options = {
    'CSV': '.csv',
    'JSON': '.json',
    'Pickle': '.pkl'
}
try:
    import openpyxl

    save_options['Excel'] = '.xlsx'
    save_options = dict(sorted(save_options.items()))
except ImportError:
    pass
try:
    import pyarrow

    save_options['Parquet'] = '.parquet'
    save_options['Feather'] = '.feather'
    save_options = dict(sorted(save_options.items()))
except ImportError:
    pass


def save_results(
        results: dict,
        data_dir: Union[str, os.PathLike[str]],
        output_format: str
) -> None:
    """Export DataFrame objects to specified directory.

    Parameters
    ----------
    results : dict
        Dictionary of pandas.DataFrame objects to be saved.
    data_dir : str or PathLike str
        Path for parent directory to save files to.
    output_format : str
        Format for saved results. Acceptable options are seen in
        save_options.

    Raises
    ------
    TypeError
        "results" not of type dict or "datadir" not of type str.
    ValueError
        Incorrect output_format provided.

    See Also
    --------
    os.PathLike : ABC for objects representing a file system path.
    save_options : Output format to extension dict.
    """

    if not isinstance(results, dict):
        raise TypeError(
            f'results must be of type dict, not \'{type(results)}\'.'
        )
    if not isinstance(data_dir, str):
        raise TypeError(
            f'data_dir must of type str, not \'{type(data_dir)}\'.'
        )

    try:
        extension = save_options[output_format]
    except KeyError:
        raise ValueError(
            f'output_format must be one of {list(save_options.keys())}, '
            f'not \'{output_format}\'.'
        )

    if not os.path.isdir(data_dir):
        os.makedirs(data_dir)

    for query, df in results.items():
        if isinstance(query, str):
            output_filename = f'{query}{extension}'
        else:
            search_term, search_type = query
            output_filename = f'{search_term}_{search_type}{extension}'

        output_filename = validate_save_filename(
            output_filename
        )

        save_dataframe(
            results=df,
            filepath=os.path.join(data_dir, output_filename),
            output_format=output_format
        )


def save_dataframe(
        results: pd.DataFrame,
        filepath: Union[str, os.PathLike[str]],
        output_format: str
) -> None:
    """Saves the specified results to the file provided.

    Parameters
    ----------
    results : pandas.DataFrame
        If DataFrame, results will be stored in a csv format.
    filepath : str or PathLike str
        Path to save file. Take note of output type as specified
        above, as appending the incorrect filetype may result in the
        file being unreadable.
    output_format : str
        Format for saved results. Acceptable options are seen in
        save_options.

    Raises
    ------
    ValueError
        If a non-dataframe object results or an incorrect output_format
        is provided.

    See Also


    Examples
    --------
    >>> data = {"id": [123, 432, 576, 194], "value": [0, 1, 1, 0]}
    >>> save_dataframe(
    ...     results=data,
    ...     filepath='data/id_vals.json',
    ...     output_format='JSON'
    ... )
    Traceback (most recent call last):
        ...
    ValueError: 'results' must be of type pandas.DataFrame, not 'dict'.

    >>> df1 = pd.DataFrame(data)
    >>> save_dataframe(
    ...     results=df1,
    ...     filepath='data/id_vals.json',
    ...     output_format='JSON'
    ... )

    >>> df2 = df1
    >>> df2.loc[:, 'Name'] = ('Jeff', 'Scott', 'Will', 'Dan')
    >>> save_dataframe(
    ...     results=df2,
    ...     filepath='data/updated_id_vals.csv',
    ...     output_format='CSV'
    ... )

    >>> save_dataframe(
    ...     results=df2,
    ...     filepath='data/full_data.txt',
    ...     output_format='TXT'
    ... )
    Traceback (most recent call last):
        ...
    ValueError: 'TXT' is not supported. 'output_format' must be one of ['CSV', 'JSON', 'Pickle'].
    """

    if not isinstance(results, pd.DataFrame):
        raise ValueError(
            f'\'results\' must be of type pandas.DataFrame, not'
            f' \'{type(results)}\'.'
        )

    try:
        getattr(results, f'to_{output_format.lower()}')(filepath)
    except AttributeError:
        raise ValueError(
            f'\'{output_format}\' is not supported. \'output_format\''
            f' must be one of {list(save_options.values())}.'
        )



import json
import os
from collections.abc import Collection
from typing import Optional, Union

import pandas as pd
from kaggle import KaggleApi
from kaggle.rest import ApiException

from pycurator._typing import (
    JSONDict,
    SearchTerm,
    SearchType,
    TermTypeResultDict
)
from pycurator.collectors import (
    BaseCollector,
    BaseTermTypeCollector
)
from pycurator.utils.validating import validate_metadata_parameters


class KaggleCollector(BaseTermTypeCollector):
    """Kaggle collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms() or
        passed in directly to search functions.
    search_types : list-like, optional (default=None)
        Data types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions to override set
        parameter.

    Notes
    -----
    For validating Kaggle requests, read the official documentation on
    authentication at https://www.kaggle.com/docs/api.
    """

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        search_types: Optional[Collection[SearchType]] = None
    ) -> None:

        super().__init__(
            repository_name='kaggle',
            search_terms=search_terms,
            search_types=search_types
        )

        self.api = KaggleApi()
        self.api.authenticate()
        self.merge_on = 'id'

    @staticmethod
    def accepts_user_credentials() -> bool:
        return False

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ('datasets', 'kernels')

    @BaseCollector._pb_indeterminate
    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
            self,
            search_term: SearchTerm,
            search_type: SearchType,
    ) -> pd.DataFrame:
        """Queries the Kaggle API for the specified search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {'datasets', 'kernels'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        # Use search type to get relevant API function
        list_queries = getattr(self.api, f'{search_type}_list')

        page_idx = 1
        search_df = pd.DataFrame()

        self._update_query_ref(page='page_idx')
        output = list_queries(search=search_term, page=page_idx)

        while output:
            if not self.continue_running:
                self.terminate()

            if search_type == 'kernels':
                output = [vars(result) for result in output]

            output_df = pd.DataFrame(output)
            output_df['page'] = page_idx

            search_df = pd.concat(
                [search_df, output_df]
            ).reset_index(drop=True)

            page_idx += 1
            self._update_query_ref(page=page_idx)
            output = list_queries(search=search_term, page=page_idx)

        if not search_df.empty:
            search_df = search_df.rename(
                columns={'id': 'datasetId', 'ref': 'id'}
            )

            if search_type == 'datasets':
                search_df = search_df.drop(columns={'viewCount', 'voteCount'})

            search_df = search_df.convert_dtypes()

        return search_df

    def _retrieve_object_json(
            self,
            object_path: str,
            data_path: Optional[str] = f'data{os.sep}'
    ) -> Union[JSONDict, None]:
        """Queries the Kaggle API for metadata JSON file.

        Parameters
        ----------
        object_path : str
        data_path : str, optional (default='data/')
            Location to save metadata to.

        Returns
        -------
        json_data : dict or None

        Raises
        ------
        kaggle.rest.ApiException
            A query was made that was unable to be fulfilled.

        See Also
        --------
        kaggle
        """

        if not self.continue_running:
            self.terminate()

        try:
            self.api.dataset_metadata(dataset=object_path, path=data_path)
        except (TypeError, ApiException) as e:
            if (isinstance(e, ApiException) and e.status != 404
                    and 'bigquery' not in e.headers['Turbolinks-Location']):
                raise e
            else:
                return None
        else:
            metadata_file_path = f'{data_path}dataset-metadata.json'
            with open(metadata_file_path) as f:
                json_data = json.load(f)

            os.remove(metadata_file_path)

            return json_data

    def get_query_metadata(
            self,
            object_paths: Collection[str],
    ) -> pd.DataFrame:
        """Retrieves the metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or list-like of str

        Returns
        -------
        metadata_df : pandas.DataFrame
        """

        object_paths = validate_metadata_parameters(object_paths)

        metadata_df = pd.DataFrame()

        for object_path in self._pb_determinate(object_paths):
            json_data = self._retrieve_object_json(object_path=object_path)
            metadata_df = pd.concat(
                [metadata_df, pd.DataFrame(json_data)]
            ).reset_index(drop=True)

        metadata_df = metadata_df.convert_dtypes()

        return metadata_df

    def get_all_metadata(
            self,
            search_dict: TermTypeResultDict
    ) -> TermTypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict of pandas.DataFrames
            Output from get_all_search_outputs function.

        Returns
        -------
        metadata_dict : dict of DataFrames
            Stores the results of each call to get_query_metadata in the
            form: metadata_dict[(search_term, search_type)] = df.
        """

        object_path_dict = dict()

        for query, df in search_dict.items():
            # Only want to get metadata for non-empty dataset DataFrames
            if 'kernels' not in query and df is not None:
                # Extract object paths
                object_paths = df.id.values
                object_path_dict[query] = object_paths

        metadata_dict = super().get_all_metadata(
            object_path_dict=object_path_dict
        )

        return metadata_dict



#
#  This software is Copyright © 2021-2022 The Regents of the University of California.

import tkinter as tk
import tkinter.ttk as ttk
from textwrap import dedent
from typing import Any, NoReturn

from .base import ViewPage
from .selection_page import SelectionPage


class LandingPage(ViewPage):
    """Landing Page of the PyCurator View component.

    Parameters
    ----------
    *args : tuple, optional
        Positional arguments for ViewPage constructor, which further
        passes to ttk.Frame.
    **kwargs : dict, optional
        Keyword arguments for ViewPage constructor, which further
        passes to ttk.Frame.

    See Also
    --------
    pycurator.gui.base.ViewPage
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @ViewPage.no_overwrite
    def show(self) -> None:
        """Arrange and display Landing Page of PyCurator UI."""
        label = ttk.Label(
            self,
            text='PyCurator',
            font='helvetica 16 bold'
        )

        message = tk.StringVar(
            value=dedent(
                """  
                Making repository research curation as easy as py.
    
                This software is Copyright © 2021-2022 
                The Regents of the University of California.
                All Rights Reserved.
                For full licensing information, reference the LICENSE present
                in the PyCurator repository.
                
                The initial development of this program was funded by the
                Librarians Association of the University of California (LAUC)
                and UC San Diego Library Research Data Curation Program (RDCP).
                """
            )
        )

        message_box = tk.Message(self, textvariable=message)

        self.next_page_button = ttk.Button(
            self,
            text='Continue',
            command=lambda: self.controller.show(SelectionPage)
        )

        label.grid(row=0, pady=7)
        message_box.grid(row=1, padx=10)
        self.next_page_button.grid(row=2, pady=5)
        self.grid(row=0, column=0, sticky='nsew')
        self.tkraise()

    def reset_frame(self) -> NoReturn:
        pass



from collections.abc import Collection
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from pycurator._typing import (
    SearchTerm,
    TermResultDict
)
from pycurator.collectors import (
    BaseCollector,
    BaseTermCollector,
)
from pycurator.utils.validating import validate_metadata_parameters


class DryadCollector(BaseTermCollector):
    """DataDryad collector for search term queries.

    Parameters
    ----------
    search_terms : list-like, optional
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        credentials: Optional[bool] = None,
    ) -> None:

        super().__init__(
            repository_name='dryad',
            search_terms=search_terms,
            credentials=credentials
        )

        self.base_url = 'https://datadryad.org/api/v2'
        self.merge_on = 'version'

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @BaseCollector._pb_indeterminate
    def _conduct_search_over_pages(
        self,
        search_url: str,
        search_params: Any,
        print_progress: bool = False,
        delim: Optional[str] = None
    ) -> pd.DataFrame:
        """Query records from the Dryad API for given parameters.

        Parameters
        ----------
        search_url : str
        search_params : dict
            Contains parameters to pass to requests.get({params}).
            Most common include search term 'q', and page index 'page'.
            For full details, see the Notes.
        print_progress : bool, optional (default=False)
            If True, updates on query page progress is sent to object
            queue to be displayed in UI window.
        delim : bool, optional (default=None)
            Key to grab results from query response JSON. If None,
            entire JSON return is considered as the data results.

        Returns
        -------
        search_df : pandas.DataFrame

        Notes
        -----
        Dryad allows the following parameters when querying the noted
        record type.

        Datasets:
        page : int, optional
            Page to search over.
        per_page : int, optional
            Number of results per page.
        q : str, optional
            Term to query for.
        affiliation : str, optional
            ROR identifier to require in a dataset's authors.
        tenant : str, optional
            Tenant organization in Dryad. Ignored if affiliation given.
        modifiedSince : str, optional
            An ISO 8601 UTC timestamp for limiting results.

        Files of a dataset version:
        id : int
            Version ID of the dataset.
        page : int, optional
            As above.
        per_page : int, optional
            As above.

        When searching for file metadata, only id, as described above,
        is allowed.
        """

        search_df = pd.DataFrame()

        if print_progress:
            self._update_query_ref(
                search_term=search_params['q'],
                page=search_params['page']
            )
        _, output = self.get_request_output(
            url=search_url,
            params=search_params
        )

        while output.get('count'):
            output = output['_embedded']

            if delim:
                output = output[delim]

            output_df = pd.DataFrame(output)
            output_df['page'] = search_params['page']

            search_df = pd.concat([
                search_df, output_df]
            ).reset_index(drop=True)

            search_params['page'] += 1
            if print_progress:
                self._update_query_ref(
                    search_term=search_params['q'],
                    page=search_params['page']
                )
            _, output = self.get_request_output(
                url=search_url,
                params=search_params
            )

        return search_df

    @BaseTermCollector.validate_search_term
    def get_individual_search_output(
            self,
            search_term: SearchTerm
    ) -> pd.DataFrame:
        """Returns information about all datasets from DataDryad.

        Parameters
        ----------
        search_term : str

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        """

        search_url = f'{self.base_url}/search'
        search_params = {'q': search_term, 'page': 1, 'per_page': 100}

        search_df = self._conduct_search_over_pages(
            search_url=search_url,
            search_params=search_params,
            print_progress=True,
            delim='stash:datasets'
        )

        # Add dataset-specific version id for metadata querying
        search_df['version'] = self._extract_version_ids(search_df)

        return search_df

    def get_query_metadata(
            self,
            object_paths: Union[str, Collection[str]],
            **kwargs: Any
    ) -> pd.DataFrame:
        """Retrieves the metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or collection of str

        Returns
        -------
        metadata_df : pandas.DataFrame

        Raises
        ------
        TypeError
            If no object paths are provided.
        """

        object_paths = validate_metadata_parameters(object_paths)

        start_page = 1
        metadata_df = pd.DataFrame()

        for object_path in self._pb_determinate(object_paths):
            search_url = f'{self.base_url}/versions/{object_path}/files'
            search_params = {'page': start_page}

            object_df = self._conduct_search_over_pages(
                search_url=search_url,
                search_params=search_params,
                delim='stash:files',
                print_progress=False
            )

            object_df['version'] = object_path
            object_df.loc[:, 'page'] = search_params['page']
            metadata_df = pd.concat(
                [metadata_df, object_df]
            ).reset_index(drop=True)

        return metadata_df

    def _extract_version_ids(self, df: pd.DataFrame) -> pd.DataFrame:
        """Retrieve ids from DataFrame entries."""
        return df['_links'].apply(
            lambda entry: entry.get('stash:version', {})
                               .get('href', '')
                               .split('/')[-1]
            if entry is not np.nan else None
        )

    def get_all_metadata(self, search_dict: TermResultDict) -> TermResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict
        """

        object_path_dict = {
            query: self._extract_version_ids(df=df)
            for query, df in search_dict.items()
        }

        metadata_dict = super().get_all_metadata(
            object_path_dict=object_path_dict
        )

        return metadata_dict



from .base import (
    BaseCollector,
    BaseAPICollector,
    BaseTermCollector,
    BaseTypeCollector,
    BaseTermTypeCollector,
    TermQueryMixin,
    TypeQueryMixin,
)
from .term_collectors import DryadCollector, ZenodoCollector
from .term_type_collectors import (
    DataverseCollector,
    FigshareCollector,
    PapersWithCodeCollector
)

available_repos = {
    'Dryad': DryadCollector,
    'Zenodo': ZenodoCollector,
    'Dataverse': DataverseCollector,
    'Figshare': FigshareCollector,
    'Papers With Code': PapersWithCodeCollector,
}

try:
    from .term_type_collectors import KaggleCollector
    available_repos['Kaggle'] = KaggleCollector
except ImportError:
    pass

try:
    from .type_collectors import OpenMLCollector
    available_repos['OpenML'] = OpenMLCollector
except ImportError:
    pass

available_repos = dict(sorted(available_repos.items()))



from .MVC import CuratorView, CuratorController, CollectorModel
from .base import PyCuratorUI, ThreadedRun, ViewPage
from .landing_page import LandingPage
from .run_page import RunPage
from .selection_page import SelectionPage



from collections.abc import Collection
from typing import Any, Optional, Union

import pandas as pd

from pycurator._typing import (
    SearchTerm,
    SearchType,
    SearchQuery,
    TermResultDict,
    TermTypeResultDict,
    TypeResultDict
)
from pycurator.collectors import (
    BaseCollector,
    BaseTermTypeCollector
)
from pycurator.utils.validating import validate_metadata_parameters


class PapersWithCodeCollector(BaseTermTypeCollector):
    """PapersWithCode collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions to override set
        parameter.
    search_types : list-like, optional
        Types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        search_types: Optional[Collection[SearchType]] = None,
        credentials: Optional[str] = None
    ) -> None:
        super().__init__(
            repository_name='paperswithcode',
            search_terms=search_terms,
            search_types=search_types,
            credentials=credentials
        )
        self.base_url = 'https://paperswithcode.com/api/v1'

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ('conferences', 'datasets', 'evaluations', 'papers', 'tasks')

    @BaseCollector._pb_indeterminate
    def _conduct_search_over_pages(
        self,
        search_url: str,
        search_params: Optional[Any] = None,
        print_progress: bool = False
    ) -> Union[pd.DataFrame, None]:
        """Query paginated results from the Papers With Code API.

        Parameters
        ----------
        search_url : str
        search_params : dict
            Contains parameters to pass to requests.get({params}). Most
            common include search term 'q', and page index 'page'. For
            full details, see the Notes.
        print_progress : bool, optional (default=False)
            If True, updates on query page progress is sent to object
            queue to be displayed in UI window.

        Returns
        -------
        search_df : pandas.DataFrame or None

        Notes
        -----
        For querying base objects, the following parameters are most common:
        q : str, optional
            Term to query for.
        page : int, optional
            Page number to request.
        items_per_page : int, optional
            Number of results to return.
        ordering : str, optional
            Which field to order results by.

        For more detailed parameters, visit
        https://paperswithcode.com/api/v1/docs/
        """

        search_df = pd.DataFrame()

        if print_progress:
            self._update_query_ref(page=search_params['page'])
        _, output = self.get_request_output(
            url=search_url,
            params=search_params
        )

        while output.get('results'):
            output = output['results']

            output_df = pd.DataFrame(output)
            output_df['page'] = search_params['page']

            search_df = pd.concat(
                [search_df, output_df]
            ).reset_index(drop=True)

            search_params['page'] += 1

            if print_progress:
                self._update_query_ref(page=search_params['page'])

            response, output = self.get_request_output(
                url=search_url,
                params=search_params
            )

            # Ensure we've received results if they exist
            # 200: OK, 404: page not found (no more results)
            while response.status_code not in [200, 404]:
                self.status_queue.put(
                    f'Search error "{response.status_code}" on '
                    f'page {search_params["page"]}'
                )
                search_params['page'] += 1

                if print_progress:
                    self._update_query_ref(page=search_params['page'])

                response, output = self.get_request_output(
                    url=search_url,
                    params=search_params
                )

        if not search_df.empty:
            return search_df
        else:
            return None

    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
            self,
            search_term: SearchTerm,
            search_type: SearchType
    ) -> pd.DataFrame:
        """Queries Papers With Code API for the specified search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {
            'conferences', 'datasets', 'evaluations', 'papers', 'tasks'
        }

        Returns
        -------
        pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        search_url = f'{self.base_url}/{search_type}'

        search_params = {
            'q': search_term,
            'page': 1,
            'items_per_page': 500
        }

        return self._conduct_search_over_pages(
            search_url=search_url,
            search_params=search_params,
            print_progress=True
        )

    @BaseTermTypeCollector.validate_search_type
    def _get_metadata_types(
            self,
            search_type: SearchType
    ) -> Collection[SearchType]:

        if search_type == 'conferences':
            return ['proceedings']
        elif search_type == 'datasets':
            return ['evaluations']
        elif search_type == 'evaluations':
            return ['metrics', 'results']
        elif search_type == 'papers':
            return ['datasets', 'methods', 'repositories', 'results', 'tasks']
        elif search_type == 'tasks':
            return ['children', 'evaluations', 'papers', 'parents']

    def get_query_metadata(
            self,
            object_paths: Collection[str],
            **kwargs: Any
    ) -> TypeResultDict:
        """Retrieves metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or collection of str
        **kwargs : dict, optional
            Must include search_type : {
                'conferences', 'datasets', 'evaluations', 'papers', 'tasks'
            }
            Search_type in kwargs to match signature of base method.

        Returns
        -------
        metadata_dict : dict
            Results are stored in the format
            metadata_dict[metadata_type] = DataFrame
        """

        try:
            search_type = kwargs['search_type']
        except KeyError:
            raise AttributeError(
                'search_type must be passed to get_query_metadata.'
            )
        object_paths = validate_metadata_parameters(object_paths)

        metadata_types = self._get_metadata_types(search_type)
        metadata_dict = dict()

        for metadata_type in metadata_types:
            search_df = pd.DataFrame()
            self.status_queue.put(f'Querying {metadata_type}.')

            for object_path in self._pb_determinate(object_paths):
                search_url = f'{self.base_url}/{search_type}/{object_path}/' \
                             f'{metadata_type}'
                search_params = {'page': 1}

                # Conduct the search and add supplementary info to DataFrame
                object_df = self._conduct_search_over_pages(
                    search_url=search_url,
                    search_params=search_params
                )

                if object_df is not None:
                    object_df['id'] = object_path
                    object_df['page'] = search_params['page']

                search_df = pd.concat(
                    [search_df, object_df]
                ).reset_index(drop=True)

            if not search_df.empty:
                metadata_dict[metadata_type] = search_df

        return metadata_dict

    def get_all_metadata(
            self,
            search_dict: TermTypeResultDict
    ) -> dict[SearchQuery, TermResultDict]:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict
        """

        metadata_dict = dict()

        for query, df in search_dict.items():
            if df is not None:
                search_term, search_type = query
                self.status_queue.put(
                    f'Querying {search_term} {search_type} metadata.'
                )

                object_paths = df.id.values

                metadata_dict[query] = self.get_query_metadata(
                    object_paths=object_paths,
                    search_type=search_type
                )

                self.status_queue.put(
                    f'{search_term} {search_type} metadata query complete.'
                )

        return metadata_dict

    def merge_search_and_metadata_dicts(
        self,
        search_dict,
        metadata_dict,
        on=None,
        left_on=None,
        right_on=None,
        **kwargs
    ):
        """Merges together search and metadata DataFrames by 'on' value.

        Parameters
        ----------
        search_dict : dict
            Dictionary of search output results.
        metadata_dict : dict
            Dictionary of metadata results.
        on : str or list-like, optional (default=None)
            Column name(s) to merge the two dicts on.
        left_on : str or list-like, optional (default=None)
            Column name(s) to merge the left dict on.
        right_on : str or list-like, optional (default=None)
            Column name(s) to merge the right dict on.
        **kwargs : dict, optional
            Allow users to add save value.

        Returns
        -------
        merged_dict : dict

        See Also
        --------
        pandas.merge
        """

        merged_dict = dict()

        for query_key, type_df_dict in metadata_dict.items():
            search_term, search_type = query_key
            search_df = search_dict[query_key]

            for metadata_type, metadata_df in type_df_dict.items():
                _search_type = f'{search_type}_{metadata_type}'
                df_all = pd.merge(
                    left=search_df,
                    right=metadata_df,
                    on='id',
                    left_on=left_on,
                    right_on=right_on,
                    how='outer',
                    suffixes=('_search', '_metadata')
                )

                merged_dict[(search_term, _search_type)] = df_all

        return merged_dict



from collections.abc import Collection
from typing import Optional, Union

import pandas as pd

from pycurator._typing import (
    SearchTerm,
    SearchType,
    TermTypeResultDict
)
from pycurator.collectors import (
    BaseCollector,
    BaseTermTypeCollector
)
from pycurator.utils.validating import validate_metadata_parameters


class FigshareCollector(BaseTermTypeCollector):
    """Figshare collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms() or
        passed in directly to search functions.
    search_types : list-like, optional (default=None)
        Data types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions to override set
        parameter.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        search_types: Optional[Collection[SearchType]] = None,
        credentials: Optional[str] = None
    ) -> None:
        super().__init__(
            repository_name='figshare',
            search_terms=search_terms,
            search_types=search_types,
            credentials=None
        )
        self.base_url = 'https://api.figshare.com/v2'
        self.merge_on = 'id'
        self.headers = dict()

        if credentials:
            self.credentials = self.load_credentials(
                credential_filepath=credentials
            )

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ('articles', 'collections', 'projects')

    def load_credentials(self, credential_filepath: str) -> Union[str, None]:
        """Load the credentials given filepath or token.

        Parameters
        ----------
        credential_filepath : str
            JSON filepath containing credentials in the form
            {repository_name}: {key}.

        Returns
        -------
        credentials : str or None
        """

        credentials = super().load_credentials(
            credential_filepath=credential_filepath
        )
        self.headers['Authorization'] = f'token {credentials}'
        return credentials

    @BaseCollector._pb_indeterminate
    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
            self,
            search_term: SearchTerm,
            search_type: SearchType,
    ) -> Union[TermTypeResultDict, None]:
        """Queries the Figshare API for the given search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {'articles', 'collections', 'projects'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        start_page = 1
        page_size = 1000
        search_year = 1950
        search_date = f'{search_year}-01-01'
        search_url = f'{self.base_url}/{search_type}'

        search_df = pd.DataFrame()
        output_df = None

        search_params = {
            'search_for': search_term,
            'published_since': search_date,
            'order_direction': 'asc',
            'page': start_page,
            'page_size': page_size
        }

        response, output = self.get_request_output_and_update_query_ref(
            url=search_url,
            params=search_params,
            headers=self.headers,
            published_since=search_date,
            page=start_page
        )

        while response.status_code == 200:
            while response.status_code == 200 and output:
                output_df = pd.DataFrame(output)
                output_df['search_page'] = search_params['page']
                output_df['publish_query'] = search_params['published_since']

                search_df = pd.concat(
                    [search_df, output_df]
                ).reset_index(drop=True)

                search_params['page'] += 1
                response, output = \
                    self.get_request_output_and_update_query_ref(
                        url=search_url,
                        params=search_params,
                        headers=self.headers,
                        published_since=search_params['published_since'],
                        page=search_params['page']
                    )

            if (output_df is not None
                    and output_df.shape[0] < search_params['page_size']):
                return search_df

            # Get new date to search
            search_date = search_df['published_date'].values[-1].split('T')[0]
            search_params['published_since'] = search_date
            search_params['page'] = start_page

            # Conduct search
            response, output = self.get_request_output_and_update_query_ref(
                url=search_url,
                params=search_params,
                headers=self.headers,
                published_since=search_params['published_since'],
                page=search_params['page']
            )

        return search_df

    def get_query_metadata(
            self,
            object_paths: Union[str, Collection[str]],
    ) -> pd.DataFrame:
        """
        Retrieves metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or list-like
            string or list of strings containing object paths.

        Returns
        -------
        metadata_df : pandas.DataFrame
            Metadata for the requested objects.
        """

        object_paths = validate_metadata_parameters(object_paths)

        metadata_df = pd.DataFrame()

        for object_path in self._pb_determinate(object_paths):
            _, json_data = self.get_request_output(
                url=object_path,
                headers=self.headers
            )

            metadata_df = pd.concat(
                [metadata_df, pd.DataFrame([json_data])]
            ).reset_index(drop=True)

        return metadata_df

    def get_all_metadata(
            self,
            search_dict: TermTypeResultDict
    ) -> TermTypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict of tuple of str to dataframe
        """

        object_path_dict = dict()

        for query, df in search_dict.items():
            if df is not None:
                _, search_type = query
                object_ids = df.id.convert_dtypes().tolist()
                object_paths = [
                    f'{self.base_url}/{search_type}/{object_id}'
                    for object_id in object_ids
                ]

                object_path_dict[query] = object_paths

        metadata_dict = super().get_all_metadata(
            object_path_dict=object_path_dict
        )

        return metadata_dict
